#pragma once

#include <memory>
#include <sstream>
#include <stdexcept>

namespace ast {

struct Constant;
struct BinOp;

struct Node {
    virtual ~Node() = default;
    virtual void accept( NodeVisitor &visitor ) = 0;
};

using NodePtr = std::unique_ptr< Node >;

struct Constant : Node {

    Constant() = default;
    explicit Constant( long val ) : _val( val ) { }

    long value() const { return _val; }
    void value( long v ) { _val = v; }

  private:
    long _val = 0;
};

struct BinOp : Node {

    BinOp( std::string op, NodePtr left, NodePtr right ) :
        _op( std::move( op ) ),
        _left( std::move( left ) ), _right( std::move( right ) )
    { }

    Node *left() { return _left.get(); }
    const Node *left() const { return _left.get(); }

    Node *right() { return _right.get(); }
    const Node *right() const { return _right.get(); }

    const std::string &op() const { return _op; }

  private:
    std::string _op;
    NodePtr _left;
    NodePtr _right;
};

struct Printer {
    Printer &run( Node &n );
};

struct Evaluator {
    Evaluator &run( Node &n );
};

} // namespace ast
